"""
Some generic utility routines for number handling and
calculating (specific) variances
"""

import itertools
import logging

import numpy
import time

from sourcefinder import extract
from sourcefinder import stats
from sourcefinder import utils
from sourcefinder.utility import containers
from sourcefinder.utility import coordinates
from sourcefinder.utility.uncertain import Uncertain
import dask.array as da
from scipy.interpolate import interp1d
import psutil
from multiprocessing import Pool
from functools import cached_property
from functools import partial
import sep

try:
    import ndimage
except ImportError:
    from scipy import ndimage
from numba import guvectorize, float32, int32


def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        if 'log_time' in kw:
            name = kw.get('log_name', method._name_.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            print('{0}  {1:2.2f} ms'.format(method.__name__, (te - ts) * 1000))
        return result
    return timed


def gather(*args):
    return list(args)


logger = logging.getLogger(__name__)

#
# Hard-coded configuration parameters; not user settable.
#
INTERPOLATE_ORDER = 1
MEDIAN_FILTER = 0  # If non-zero, apply a median filter of size
# MEDIAN_FILTER to the background and RMS grids prior
# to interpolating.
MF_THRESHOLD = 0  # If MEDIAN_FILTER is non-zero, only use the filtered
# grid when the (absolute) difference between the raw
# and filtered grids is larger than MF_THRESHOLD.
DEBLEND_MINCONT = 0.005  # Min. fraction of island flux in deblended subisland
STRUCTURING_ELEMENT = [[1, 1, 1], [1, 1, 1], [1, 1, 1]]  # Island connectiivty

# Let's retain the option of calculating background maps in the original way,
# i.e. without using sep. This is slower, but more accurate. Ultimately, the
# switch for choosing either should be propagated through an argument of the 
# ImageData class instantiation, but we will implement that later. For now,
# we will not use SEP as default.
SEP = False
# Vectorized processing of source islands is much faster, but excludes Gaussian
# fits, therefore slightly less accurate.
VECTORIZED = False


class ImageData(object):
    """Encapsulates an image in terms of a numpy array + meta/headerdata.

    This is your primary contact point for interaction with images: it icludes
    facilities for source extraction and measurement, etc.
    """

    def __init__(self, data, beam, wcs, margin=0, radius=0, back_size_x=32,
                 back_size_y=32, residuals=False, islands=False):
        """Sets up an ImageData object.

        *Args:*
          - data (2D numpy.ndarray): actual image data
          - wcs (utility.coordinates.wcs): world coordinate system
            specification
          - beam (3-tuple): beam shape specification as
            (semimajor, semiminor, theta)

        """

        # Do data, wcs and beam need deepcopy?
        # Probably not (memory overhead, in particular for data),
        # but then the user shouldn't change them outside ImageData in the
        # meantime
        # self.rawdata is a 2D numpy array, C-contiguous needed for sep.
        # single precision is good enough in all cases.
        self.rawdata = numpy.ascontiguousarray(data, dtype=numpy.float32)
        self.wcs = wcs  # a utility.coordinates.wcs instance
        self.beam = beam  # tuple of (semimaj, semimin, theta) in pixel coordinates.
        # These three quantities are only dependent on the beam, so should be calculated
        # once the beam is known and not for each source separately.
        self.fudge_max_pix_factor = utils.fudge_max_pix(beam[0], beam[1], beam[2])
        self.max_pix_variance_factor = utils.maximum_pixel_method_variance(
                           beam[0], beam[1], beam[2])
        self.beamsize = utils.calculate_beamsize(beam[0], beam[1])
        self.correlation_lengths = utils.calculate_correlation_lengths(beam[0], beam[1])
        self.clip = {}
        self.labels = {}
        self.freq_low = 1
        self.freq_high = 1

        self.back_size_x = back_size_x
        self.back_size_y = back_size_y
        self.margin = margin
        self.radius = radius
        self.residuals = residuals
        self.islands = islands

    ###########################################################################
    #                                                                         #
    # Properties and attributes.                                              #
    #                                                                         #
    # Properties are attributes managed by methods; rather than calling the   #
    # method directly, the attribute automatically invokes it. Result of the  #
    # function calls are cached, subsequent calls doesn't recompute.          #
    #                                                                         #
    # clearcache() clears all the cached data, which can get quite large.     #
    # It may be wise to call this, for example, in an exception handler       #
    # dealing with MemoryErrors.                                              #
    #                                                                         #
    ###########################################################################
    @cached_property
    def grids(self):
        """Gridded RMS and background data for interpolating"""
        return self.__grids()

    @cached_property
    def background(self):
        """"Returns background object from sep"""
        return sep.Background(self.data.data, mask = self.data.mask,
                              bw=self.back_size_x, bh=self.back_size_y, fw=0, fh=0)

    @cached_property
    @timeit
    def backmap(self):
        """Background map"""
        if not hasattr(self, "_user_backmap"):
            if SEP:
                return numpy.ma.array(self.background.back(), mask=self.data.mask)
            else:
                return self._interpolate(self.grids['bg'])
        else:
            return self._user_backmap

    @cached_property
    @timeit
    def rmsmap(self):
        """RMS map"""
        if not hasattr(self, "_user_noisemap"):
            if SEP:
                return numpy.ma.array(self.background.rms(), mask=self.data.mask)
            else:
                return self._interpolate(self.grids['rms'], roundup=True)
        else:
            return self._user_noisemap

    @cached_property
    def data(self):
        """Masked image data"""
        # We will ignore all the data which is masked for the rest of the
        # sourcefinding process. We build up the mask by stacking ("or-ing
        # together") a number of different effects:
        #
        # * A margin from the edge of the image;
        # * Any data outside a given radius from the centre of the image;
        # * Data which is "obviously" bad (equal to 0 or NaN).
        mask = numpy.zeros((self.xdim, self.ydim))
        if self.margin:
            margin_mask = numpy.ones((self.xdim, self.ydim))
            margin_mask[self.margin:-self.margin, self.margin:-self.margin] = 0
            mask = numpy.logical_or(mask, margin_mask)
        if self.radius:
            radius_mask = utils.circular_mask(self.xdim, self.ydim, self.radius)
            mask = numpy.logical_or(mask, radius_mask)
        mask = numpy.logical_or(mask, numpy.isnan(self.rawdata))
        return numpy.ma.array(self.rawdata, mask=mask)

    @cached_property
    def data_bgsubbed(self):
        """Background subtracted masked image data"""
        return self.data - self.backmap

    @property
    def xdim(self):
        """X pixel dimension of (unmasked) data"""
        return self.rawdata.shape[0]

    @property
    def ydim(self):
        """Y pixel dimension of (unmasked) data"""
        return self.rawdata.shape[1]

    @property
    def pixmax(self):
        """Maximum pixel value (pre-background subtraction)"""
        return self.data.max()

    @property
    def pixmin(self):
        """Minimum pixel value (pre-background subtraction)"""
        return self.data.min()

    def clearcache(self):
        """Zap any calculated data stored in this object.

        Clear the background and rms maps, labels, clip, and any locally held
        data. All of these can be reconstructed from the data accessor.

        Note that this *must* be run to pick up any new settings.
        """
        try:
            self.labels.clear()
            self.clip.clear()
            del self.background
            del self.backmap
            del self.rmsmap
            del self.data
            del self.data_bgsubbed
            del self.grids
            del self.Gaussian_islands
            del self.Gaussian_residuals
            del self.residuals_from_deblending
        except AttributeError:
            pass

    ###########################################################################
    #                                                                         #
    # General purpose image handling.                                         #
    #                                                                         #
    # Routines for saving and trimming data, and calculating background/RMS   #
    # maps (in conjuntion with the properties above).                         #
    #                                                                         #
    ###########################################################################

    # Private "support" methods
    def __grids(self):
        """Calculate background and RMS grids of this image.

        These grids can be interpolated up to make maps of the original image
        dimensions: see _interpolate().

        This is called automatically when ImageData.backmap,
        ImageData.rmsmap or ImageData.fdrmap is first accessed.
        """

        # there's no point in working with the whole of the data array
        # if it's masked.
        useful_chunk = ndimage.find_objects(numpy.where(self.data.mask, 0, 1))
        assert (len(useful_chunk) == 1)
        y_dim = self.data[useful_chunk[0]].data.shape[1]
        useful_data = da.from_array(self.data[useful_chunk[0]].data, chunks=(self.back_size_x, y_dim))

        mode_and_rms = useful_data.map_blocks(ImageData.compute_mode_and_rms_of_row_of_subimages,
                                              y_dim,  self.back_size_y,
                                              dtype=numpy.complex64,
                                              chunks=(1, 1)).compute()

        # See also similar comment below. This solution was chosen because map_blocks does not seem to be able to
        # output multiple arrays. One can however output to a complex array and take real and imaginary
        # parts afterward. Not a very clean solution, I admit.
        mode_grid = mode_and_rms.real
        rms_grid = mode_and_rms.imag

        rms_grid = numpy.ma.array(
            rms_grid, mask=numpy.where(rms_grid == 0, 1, 0))
        # A rms of zero is not physical, since any instrument has system noise, so I use that as criterion
        # to mask values. A zero background mode is physically possible, but also highly unlikely, given the way
        # we determine it.
        mode_grid = numpy.ma.array(
            mode_grid, mask=numpy.where(rms_grid == 0, 1, 0))

        return { 'bg': mode_grid, 'rms': rms_grid,}

    @staticmethod
    def compute_mode_and_rms_of_row_of_subimages(row_of_subimages, y_dim, back_size_y):

        # We set up a dedicated logging subchannel, as the sigmaclip loop
        # logging is very chatty:
        sigmaclip_logger = logging.getLogger(__name__ + '.sigmaclip')
        row_of_complex_values = numpy.empty(0, numpy.complex64)

        for starty in range(0, y_dim, back_size_y):
            chunk = row_of_subimages[:, starty:starty+back_size_y]
            if not chunk.any():
                # In the original code we had rmsrow.append(False), but now we work with an array instead of a list,
                # so I'll set these values to zero instead and use these zeroes to create the mask.
                rms = 0
                mode = 0
            else:
                chunk, sigma, median, num_clip_its = stats.sigma_clip(
                    chunk.ravel())
                if len(chunk) == 0 or not chunk.any():
                    rms = 0
                    mode = 0
                else:
                    mean = numpy.mean(chunk)
                    rms = sigma
                    # In the case of a crowded field, the distribution will be
                    # skewed and we take the median as the background level.
                    # Otherwise, we take 2.5 * median - 1.5 * mean. This is the
                    # same as SExtractor: see discussion at
                    # <http://terapix.iap.fr/forum/showthread.php?tid=267>.
                    # (mean - median) / sigma is a quick n' dirty skewness
                    # estimator devised by Karl Pearson.
                    if numpy.fabs(mean - median) / sigma >= 0.3:
                        sigmaclip_logger.debug(
                            'bg skewed, %f clipping iterations', num_clip_its)
                        mode=median
                    else:
                        sigmaclip_logger.debug(
                            'bg not skewed, %f clipping iterations',
                            num_clip_its)
                        mode=2.5 * median - 1.5 * mean
            row_of_complex_values = numpy.append(row_of_complex_values,  numpy.array(mode + 1j*rms))[None]
        # This solution is a bit dirty. I would like dask.array.map_blocks to output two arrays,
        # but presently that module does not seem to provide for that. But I can, however, output to a
        # complex array and later take the real part of that for the mode and the imaginary part
        # for the rms.
        return row_of_complex_values

    def _interpolate(self, grid, roundup=False):
        """
        Interpolate a grid to produce a map of the dimensions of the image.

        Args:

            grid (numpy.ma.MaskedArray)

        Kwargs:

            roundup (bool)

        Returns:

            (numpy.ma.MaskedArray)

        Used to transform the RMS, background or FDR grids produced by
        L{_grids()} to a map we can compare with the image data.

        If roundup is true, values of the resultant map which are lower than
        the input grid are trimmed.
        """
        # there's no point in working with the whole of the data array if it's
        # masked.
        useful_chunk = ndimage.find_objects(numpy.where(self.data.mask, 0, 1))
        assert (len(useful_chunk) == 1)
        my_xdim, my_ydim = self.data[useful_chunk[0]].shape

        if MEDIAN_FILTER:
            f_grid = ndimage.median_filter(grid, MEDIAN_FILTER)
            if MF_THRESHOLD:
                grid = numpy.where(
                    numpy.fabs(f_grid - grid) > MF_THRESHOLD, f_grid, grid
                )
            else:
                grid = f_grid

        # Bicubic spline interpolation
        xratio = float(my_xdim) / self.back_size_x
        yratio = float(my_ydim) / self.back_size_y

        my_map = numpy.ma.MaskedArray(numpy.zeros(self.data.shape),
                                      mask=self.data.mask)

        # Remove the MaskedArrayFutureWarning warning and keep old numpy < 1.11
        # behavior
        my_map.unshare_mask()

        # Inspired by https://stackoverflow.com/questions/13242382/resampling-a-numpy-array-representing-an-image
        # Should be much faster than scipy.ndimage.map_coordinates.
        # scipy.ndimage.zoom should also be an option for speedup, but zoom dit not let me produce the exact
        # same output as map_coordinates. My bad.
        # I checked, using fitsdiff, that it gives the exact same output as the original code
        # up to and including --relative-tolerance=1e-15 for INTERPOLATE_ORDER=1.
        # It was actually quite a hassle to get the same output and the fill_value is essential
        # in interp1d. However, for some unit tests, grid.shape=(1,1) and then it will break
        # with "ValueError: x and y arrays must have at least 2 entries". So in that case
        # map_coordinates should be used.

        if INTERPOLATE_ORDER==1 and grid.shape[0]>1 and grid.shape[1]>1:
            x_initial = numpy.linspace(0., grid.shape[0]-1, grid.shape[0], endpoint=True)
            y_initial = numpy.linspace(0., grid.shape[1]-1, grid.shape[1], endpoint=True)
            x_sought = numpy.linspace(-0.5, -0.5 + xratio, my_xdim, endpoint=True)
            y_sought = numpy.linspace(-0.5, -0.5 + yratio, my_ydim, endpoint=True)

            primary_interpolation = interp1d(y_initial, grid, kind='slinear', assume_sorted=True,
                                             axis=1, copy=False, bounds_error=False,
                                             fill_value=(grid[:, 0], grid[:, -1]))
            transposed = primary_interpolation(y_sought).T

            perpendicular_interpolation = interp1d(x_initial, transposed, kind='slinear', assume_sorted=True,
                                                   axis=1, copy=False, bounds_error=False,
                                                   fill_value=(transposed[:, 0], transposed[:, -1]))
            my_map[useful_chunk[0]] = perpendicular_interpolation(x_sought).T
        else:
            slicex = slice(-0.5, -0.5 + xratio, 1j * my_xdim)
            slicey = slice(-0.5, -0.5 + yratio, 1j * my_ydim)
            my_map[useful_chunk[0]] = ndimage.map_coordinates(
               grid, numpy.mgrid[slicex, slicey],
               mode='nearest', order=INTERPOLATE_ORDER)

        # If the input grid was entirely masked, then the output map must
        # also be masked: there's no useful data here. We don't search for
        # sources on a masked background/RMS, so this data will be cleanly
        # skipped by the rest of the sourcefinder
        if numpy.ma.getmask(grid).all():
            my_map.mask = True
        elif roundup:
            # In some cases, the spline interpolation may produce values
            # lower than the minimum value in the map. If required, these
            # can be trimmed off. No point doing this if the map is already
            # fully masked, though.
            my_map = numpy.ma.MaskedArray(
                data=numpy.where(
                    my_map >= numpy.min(grid), my_map, numpy.min(grid)),
                mask=my_map.mask
            )
        return my_map

    ###########################################################################
    #                                                                         #
    # Source extraction.                                                      #
    #                                                                         #
    # Provides for both traditional (islands-above-RMS) and FDR source        #
    # extraction systems.                                                     #
    #                                                                         #
    ###########################################################################

    def extract(self, det, anl, noisemap=None, bgmap=None, labelled_data=None,
                labels=None, deblend_nthresh=0, force_beam=False):

        """
        Kick off conventional (ie, RMS island finding) source extraction.

        Kwargs:

            det (float): detection threshold, as a multiple of the RMS
                noise. At least one pixel in a source must exceed this
                for it to be regarded as significant.

            anl (float): analysis threshold, as a multiple of the RMS
                noise. All the pixels within the island that exceed
                this will be used when fitting the source.

            noisemap (numpy.ndarray):

            bgmap (numpy.ndarray):

            deblend_nthresh (int): number of subthresholds to use for
                deblending. Set to 0 to disable.

            force_beam (bool): force all extractions to have major/minor axes
                equal to the restoring beam

        Returns:
             :class:`sourcefinder.utility.containers.ExtractionResults`
        """

        if anl > det:
            logger.warning(
                "Analysis threshold is higher than detection threshold"
            )

        # If the image data is flat we may as well crash out here with a
        # sensible error message, otherwise the RMS estimation code will
        # crash out with a confusing error later.
        if numpy.ma.max(self.data) == numpy.ma.min(self.data):
            raise RuntimeError("Bad data: Image data is flat")

        if (type(bgmap).__name__ == 'ndarray' or
                type(bgmap).__name__ == 'MaskedArray'):
            if bgmap.shape != self.backmap.shape:
                raise IndexError("Background map has wrong shape")
            else:
                self.backmap = bgmap

        if (type(noisemap).__name__ == 'ndarray' or
                                       type(noisemap).__name__ == 'MaskedArray'):
            if noisemap.shape != self.rmsmap.shape:
                raise IndexError("Noisemap has wrong shape")
            if noisemap.min() < 0:
                raise ValueError("RMS noise cannot be negative")
            else:
                self.rmsmap = noisemap

        if labelled_data is not None and labelled_data.shape != self.data.shape:
            raise ValueError("Labelled map is wrong shape")

        return self._pyse(
            det * self.rmsmap, anl * self.rmsmap, deblend_nthresh, force_beam,
            labelled_data=labelled_data, labels=labels
        )

    def reverse_se(self, det, anl):
        """Run source extraction on the negative of this image.

        Obviously, there should be no sources in the negative image, so this
        tells you about the false positive rate.

        We need to clear cached data -- backgroung map, cached clips, etc --
        before & after doing this, as they'll interfere with the normal
        extraction process. If this is regularly used, we'll want to
        implement a separate cache.
        """
        self.labels.clear()
        self.clip.clear()
        self.data_bgsubbed *= -1
        results = self.extract(det=det, anl=anl)
        self.data_bgsubbed *= -1
        self.labels.clear()
        self.clip.clear()
        return results

    def fd_extract(self, alpha, anl=None, noisemap=None,
                   bgmap=None, deblend_nthresh=0, force_beam=False
                   ):
        """False Detection Rate based source extraction.
        The FDR procedure guarantees that <FDR> < alpha.

        See `Hopkins et al., AJ, 123, 1086 (2002)
        <http://adsabs.harvard.edu/abs/2002AJ....123.1086H>`_.
        """

        # The correlation length in config.py is used not only for the
        # calculation of error bars with the Condon formulae, but also for
        # calculating the number of independent pixels.
        corlengthlong, corlengthshort = self.correlation_lengths

        C_n = (1.0 / numpy.arange(
            round(0.25 * numpy.pi * corlengthlong *
                  corlengthshort + 1))[1:]).sum()

        # Calculate the FDR threshold
        # Things will go terribly wrong in the line below if the interpolated
        # noise values get very close or below zero. Use INTERPOLATE_ORDER=1
        # or the roundup option.
        if (type(bgmap).__name__ == 'ndarray' or
                    type(bgmap).__name__ == 'MaskedArray'):
            if bgmap.shape != self.backmap.shape:
                raise IndexError("Background map has wrong shape")
            else:
                self.backmap = bgmap
        if (type(noisemap).__name__ == 'ndarray' or
                    type(noisemap).__name__ == 'MaskedArray'):
            if noisemap.shape != self.rmsmap.shape:
                raise IndexError("Noisemap has wrong shape")
            if noisemap.min() < 0:
                raise ValueError("RMS noise cannot be negative")
            else:
                self.rmsmap = noisemap

        normalized_data = self.data_bgsubbed / self.rmsmap

        n1 = numpy.sqrt(2 * numpy.pi)
        prob = numpy.sort(
            numpy.ravel(numpy.exp(-0.5 * normalized_data ** 2) / n1))
        lengthprob = float(len(prob))
        compare = (alpha / C_n) * numpy.arange(lengthprob + 1)[1:] / lengthprob
        # Find the last undercrossing, see, e.g., fig. 9 in Miller et al., AJ
        # 122, 3492 (2001).  Searchsorted is not used because the array is not
        # sorted.
        try:
            index = (numpy.where(prob - compare < 0.)[0]).max()
        except ValueError:
            # Everything below threshold
            return containers.ExtractionResults()

        fdr_threshold = numpy.sqrt(-2.0 * numpy.log(n1 * prob[index]))
        # Default we require that all source pixels are above the threshold,
        # not only the peak pixel.  This gives a better guarantee that indeed
        # the fraction of false positives is less than fdr_alpha in config.py.
        # See, e.g., Hopkins et al., AJ 123, 1086 (2002).
        if not anl:
            anl = fdr_threshold
        return self._pyse(fdr_threshold * self.rmsmap, anl * self.rmsmap,
                          deblend_nthresh, force_beam)

    def flux_at_pixel(self, x, y, numpix=1):
        """Return the background-subtracted flux at a certain position
        in the map"""

        # numpix is the number of pixels to look around the target.
        # e.g. numpix = 1 means a total of 9 pixels, 1 in each direction.
        return self.data_bgsubbed[y - numpix:y + numpix + 1,
               x - numpix:x + numpix + 1].max()

    @staticmethod
    def box_slice_about_pixel(x, y, box_radius):
        """
        Returns a slice centred about (x,y), of width = 2*int(box_radius) + 1
        """
        ibr = int(box_radius)
        x = int(x)
        y = int(y)
        return (slice(x - ibr, x + ibr + 1),
                slice(y - ibr, y + ibr + 1))

    def fit_to_point(self, x, y, boxsize, threshold, fixed):
        """Fit an elliptical Gaussian to a specified point on the image.

        The fit is carried on a square section of the image, of length
        *boxsize* & centred at pixel coordinates *x*, *y*. Any data
        below *threshold* * rmsmap is not used for fitting. If *fixed*
        is set to ``position``, then the pixel coordinates are fixed
        in the fit.

        Returns an instance of :class:`sourcefinder.extract.Detection`.
        """

        logger.debug("Force-fitting pixel location ({},{})".format(x, y))
        # First, check that x and y are actually valid semi-positive integers.
        # Otherwise,
        # If they are too high (positive), then indexing will fail
        # BUT, if they are negative, then we get wrap-around indexing
        # and the fit continues at the wrong position!
        if (x < 0 or x > self.xdim
            or y < 0 or y > self.ydim):
            logger.warning("Dropping forced fit at ({},{}), "
                           "pixel position outside image".format(x, y)
                           )
            return None

        # Next, check if any of the central pixels (in a 3x3 box about the
        # fitted pixel position) have been Masked
        # (e.g. if NaNs, or close to image edge) - reject if so.
        central_pixels_slice = ImageData.box_slice_about_pixel(x, y, 1)
        if self.data.mask[central_pixels_slice].any():
            logger.warning(
                "Dropping forced fit at ({},{}), "
                "Masked pixel in central fitting region".format(x, y))
            return None

        if ((
                    # Recent NumPy
                    hasattr(numpy.ma.core, "MaskedConstant") and
                    isinstance(self.rmsmap, numpy.ma.core.MaskedConstant)
            ) or (
                # Old NumPy
                numpy.ma.is_masked(self.rmsmap[int(x), int(y)])
        )):
            logger.error("Background is masked: cannot fit")
            return None

        chunk = ImageData.box_slice_about_pixel(x, y, boxsize / 2.0)
        if threshold is not None:
            # We'll mask out anything below threshold*self.rmsmap from the fit.
            labels, num = self.labels.setdefault(
                # Dictionary mapping threshold -> islands map
                threshold,
                ndimage.label(
                    self.clip.setdefault(
                        # Dictionary mapping threshold -> mask
                        threshold,
                        numpy.where(
                            self.data_bgsubbed > threshold * self.rmsmap, 1, 0
                        )
                    )
                )
            )

            mylabel = labels[int(x), int(y)]
            if mylabel == 0:  # 'Background'
                raise ValueError(
                    "Fit region is below specified threshold, fit aborted.")
            mask = numpy.where(labels[chunk] == mylabel, 0, 1)
            fitme = numpy.ma.array(self.data_bgsubbed[chunk], mask=mask)
            if len(fitme.compressed()) < 1:
                raise IndexError("Fit region too close to edge or too small")
        else:
            fitme = self.data_bgsubbed[chunk]
            if fitme.size < 1:
                raise IndexError("Fit region too close to edge or too small")

        if not len(fitme.compressed()):
            logger.error("All data is masked: cannot fit")
            return None

        # set argument for fixed parameters based on input string
        if fixed == 'position':
            fixed = {'xbar': boxsize / 2.0, 'ybar': boxsize / 2.0}
        elif fixed == 'position+shape':
            fixed = {'xbar': boxsize / 2.0, 'ybar': boxsize / 2.0,
                     'semimajor': self.beam[0],
                     'semiminor': self.beam[1],
                     'theta': self.beam[2]}
        elif fixed is None:
            fixed = {}
        else:
            raise TypeError("Unkown fixed parameter")

        if threshold is not None:
            threshold_at_pixel = threshold * self.rmsmap[int(x), int(y)]
        else:
            threshold_at_pixel = None

        try:
            measurement, _, _ = extract.source_profile_and_errors(
                fitme, threshold_at_pixel, self.rmsmap[int(x), int(y)],
                self.beam, self.fudge_max_pix_factor,
                self.max_pix_variance_factor,
                self.beamsize, self.correlation_lengths, fixed=fixed)
        except ValueError:
            # Fit failed to converge
            # Moments are not applicable when holding parameters fixed
            logger.error("Gaussian fit failed at %f, %f", x, y)
            return None

        try:
            assert (abs(measurement['xbar']) < boxsize)
            assert (abs(measurement['ybar']) < boxsize)
        except AssertionError:
            logger.warning('Fit falls outside of box.')

        measurement['xbar'] += x - boxsize / 2.0
        measurement['ybar'] += y - boxsize / 2.0
        measurement.sig = (fitme / self.rmsmap[chunk]).max()

        return extract.Detection(measurement, self)

    def fit_fixed_positions(self, positions, boxsize, threshold=None,
                            fixed='position+shape',
                            ids=None):
        """
        Convenience function to fit a list of sources at the given positions

        This function wraps around fit_to_point().

        Args:
            positions (tuple): list of (RA, Dec) tuples. Positions to be fit,
                in decimal degrees.
            boxsize: See :py:func:`fit_to_point`
            threshold: as above.
            fixed: as above.
            ids (tuple): A list of identifiers. If not None, then must match
                the length and order of the ``requested_fits``. Any
                successfully fit positions will be returned in a tuple
                along with the matching id. As these are simply passed back to
                calling code they can be a string, tuple or whatever.

        In particular, boxsize is in pixel coordinates as in
        fit_to_point, not in sky coordinates.

        Returns:
            tuple: A list of successful fits.
                If ``ids`` is None, returns a single list of
                :class:`sourcefinder.extract.Detection` s.
                Otherwise, returns a tuple of two matched lists:
                ([detections], [matching_ids]).
        """

        if ids is not None:
            assert len(ids) == len(positions)

        successful_fits = []
        successful_ids = []
        for idx, posn in enumerate(positions):
            try:
                x, y, = self.wcs.s2p((posn[0], posn[1]))
            except RuntimeError as e:
                if (str(e).startswith("wcsp2s error: 8:") or
                        str(e).startswith("wcsp2s error: 9:")):
                    logger.warning("Input coordinates (%.2f, %.2f) invalid: ",
                                   posn[0], posn[1])
                else:
                    raise
            else:
                try:
                    fit_results = self.fit_to_point(x, y,
                                                    boxsize=boxsize,
                                                    threshold=threshold,
                                                    fixed=fixed)
                    if not fit_results:
                        # We were unable to get a good fit
                        continue
                    if (fit_results.ra.error == float('inf') or
                            fit_results.dec.error == float('inf')):
                        logging.warning("position errors extend outside image")
                    else:
                        successful_fits.append(fit_results)
                        if ids:
                            successful_ids.append(ids[idx])

                except IndexError as e:
                    logger.warning("Input pixel coordinates (%.2f, %.2f) "
                                   "could not be fit because: " + str(e),
                                   posn[0], posn[1])
        if ids:
            return successful_fits, successful_ids
        return successful_fits

    def label_islands(self, detectionthresholdmap, analysisthresholdmap, deblend_nthresh):
        """
        Return a lablled array of pixels for fitting.

        Args:

            detectionthresholdmap (numpy.ndarray):

            analysisthresholdmap (numpy.ndarray):

            deblend_nthresh: number of thresholds for deblending (integer)

        Returns:

            list of valid islands (list of int)

            labelled islands (numpy.ndarray)
        """
        # If there is no usable data, we return an empty set of islands.
        if not len(self.rmsmap.compressed()):
            logging.warning("RMS map masked; sourcefinding skipped")
            return [], numpy.zeros(self.data_bgsubbed.shape, dtype=numpy.int)

        # At this point, we select all the data which is eligible for
        # sourcefitting. We are actually using three separate filters, which
        # exclude:
        #
        # 1. Anything which has been masked before we reach this point;
        # 2. Any pixels which fall below the analysis threshold at that pixel
        #    position;
        # 3. Any pixels corresponding to a position where the RMS noise is
        #    less than RMS_FILTER (default 0.001) times the median RMS across
        #    the whole image.
        #
        # The third filter attempts to exclude those regions of the image
        # which contain no usable data; for example, the parts of the image
        # falling outside the circular region produced by awimager.
        RMS_FILTER = 0.001
        # combined_mask = numpy.logical_or(self.rmsmap.data < RMS_FILTER * self.background.globalrms,
        #                                 self.data.mask)

        clipped_data = numpy.ma.where(
            (self.data_bgsubbed > analysisthresholdmap) &
            (self.rmsmap >= (RMS_FILTER * self.background.globalrms)),
            1, 0
        ).filled(fill_value=0)
        labelled_data, num_labels = ndimage.label(clipped_data,
                                                  STRUCTURING_ELEMENT)

        # Get a bounding box for each island:
        # NB Slices ordered by label value (1...N,)
        # 'None' returned for missing label indices.
        slices = ndimage.find_objects(labelled_data)

        # Derive all indices than correspond to the slices
        all_indices = ImageData.slices_to_indices(slices)

        # Derive maximum positions, maximum values and number of pixels
        # per island.
        maxposs = numpy.empty((num_labels, 2), dtype=numpy.int32)
        dummy = numpy.empty_like(maxposs)
        maxis = numpy.empty(num_labels, dtype=numpy.float32)
        npixs = numpy.empty(num_labels, dtype=numpy.int32)

        ImageData.extract_parms_image_slice(self.data_bgsubbed.data.astype(
                                            dtype=numpy.float32, copy=False),
                                            all_indices, labelled_data,
                                            numpy.arange(1, num_labels+1,
                                            dtype=numpy.int32),
                                            dummy, maxposs, maxis, npixs)

        # Here we remove the labels that correspond to islands below the
        # detection threshold.
        above_det_thr = maxis > detectionthresholdmap[maxposs[:, 0],
                                                      maxposs[:, 1]]

        num_islands_above_detection_threshold = above_det_thr.sum()

        # numpy.arange(1, num_labels + 1)) to discard the zero label
        # (background).
        # Will break in the pathological case that all the image pixels
        # are covered by sources, but we will take that risk.
        labels_above_det_thr = numpy.extract(above_det_thr, numpy.arange(1,
                                             num_labels + 1))

        maxposs_above_det_thr = numpy.compress(above_det_thr, maxposs, axis=0)
        maxis_above_det_thr = numpy.extract(above_det_thr, maxis)
        npixs_above_det = numpy.extract(above_det_thr, npixs)
        all_indices_above_det_thr = numpy.compress(above_det_thr, all_indices,
                                                   axis=0)

        print(f"Number of sources = {num_islands_above_detection_threshold}")

        return (labels_above_det_thr, labelled_data,
                num_islands_above_detection_threshold, maxposs_above_det_thr,
                maxis_above_det_thr, npixs_above_det, all_indices_above_det_thr,
                slices)

    @staticmethod
    def fit_islands(fudge_max_pix_factor, max_pix_variance_factor, beamsize,
                    correlation_lengths, fixed, island):
        return island.fit(fudge_max_pix_factor, max_pix_variance_factor, beamsize,
                          correlation_lengths, fixed=fixed)

    @staticmethod
    def slices_to_indices(slices):
        all_indices = numpy.empty((len(slices), 4), dtype=numpy.int32)
        for i in range(len(slices)):
            some_slice = slices[i]
            all_indices[i, :] = numpy.array([some_slice[0].start,
                                             some_slice[0].stop,
                                             some_slice[1].start,
                                             some_slice[1].stop])
        return all_indices

    @staticmethod
    @guvectorize([(float32[:, :], int32[:], int32[:, :], int32[:], int32[:],
                 int32[:], float32[:], int32[:])], '(n, m), (l), (n, m), ' +
                 '(), (k) -> (k), (), ()')
    def extract_parms_image_slice(some_image, inds, labelled_data, label,
                                  dummy, maxpos, maxi, npix):
        """
        For an island, indicated by a group of pixels with the same label,
        find the highest pixel value and its position, first relative to the
        upper left corner of the rectangular slice encompassing the island,
        but finally relative to the upper left corner of the image, i.e. the
        [0, 0] position of the Numpy array with all the image pixel values.
        Also, derive the number of pixels of the island.

        :param some_image: The 2D ndarray with all the pixel values, typically
                           self.data_bgsubbed.data.

        :param inds: A ndarray of four indices indicating the slice encompassing
                     an island. Such a slice would typically be a pick from a
                     list of slices from a call to scipy.ndimage.find_objects.
                     Since we are attempting vectorized processing here, the
                     slice should have been replaced by its four coordinates
                     through a call to slices_to_indices.

        :param labelled_data: A ndarray with the same shape as some_image, with
                              labelled islands with integer values and zeroes
                              for all background pixels.

        :param label: The label (integer value) corresponding to the slice
                      encompassing the island. Or actually it should be the
                      other way round, since there can be multiple islands
                      within one rectangular slice.

        :param dummy: Artefact of the implementation of guvectorize:
                      Empty array with the same shape as maxpos. It is needed
                      because of a missing feature in guvectorize: There is no
                      other way to tell guvectorize what the shape of the
                      output array will be. Therefore, we define an otherwise
                      redundant input array with the same shape as the desired
                      output array. Defined as int32, but could be any type.

        :param maxpos: Ndarray of two integers indicating the indices of the
                       highest pixel value of the island with label = label
                       relative to the position of pixel [0, 0] of the image.

        :param maxi:  Float32 equal to the highest pixel value
                      of the island with label = label.

        :param npix:  Integer indicating the number of pixels of the island.

        :return:      No return values, because of the use of the guvectorize
                      decorator: 'guvectorize() functions don’t return their
                      result value: they take it as an array argument,
                      which must be filled in by the function'. In this case
                      maxpos, maxi and npix will be filled with values.
        """

        labelled_data_chunk = labelled_data[inds[0]:inds[1], inds[2]:inds[3]]
        image_chunk = some_image[inds[0]:inds[1], inds[2]:inds[3]]
        segmented_island = numpy.where(labelled_data_chunk == label[0], 1, 0)

        selected_data = segmented_island * image_chunk
        maxpos_flat = selected_data.argmax()
        maxpos[0] = numpy.floor_divide(maxpos_flat, selected_data.shape[1])
        maxpos[1] = numpy.mod(maxpos_flat, selected_data.shape[1])
        maxi[0] = selected_data[maxpos[0], maxpos[1]]
        maxpos[0] += inds[0]
        maxpos[1] += inds[2]
        npix[0] = int32(segmented_island.sum())

    @timeit
    def _pyse(
            self, detectionthresholdmap, analysisthresholdmap,
            deblend_nthresh, force_beam, labelled_data=None, labels=[],
            eps_ra=0, eps_dec=0
    ):
        """
        Run Python-based source extraction on this image.

        Args:

            detectionthresholdmap (numpy.ndarray):

            analysisthresholdmap (numpy.ndarray):

            deblend_nthresh (int): number of subthresholds for deblending. 0
                disables.

            force_beam (bool): force all extractions to have major/minor axes
                equal to the restoring beam

            labelled_data (numpy.ndarray): labelled island map (output of
            numpy.ndimage.label()). Will be calculated automatically if not
            provided.

            labels (tuple): list of labels in the island map to use for
            fitting.

        Returns:

            (.utility.containers.ExtractionResults):

        This is described in detail in the "Source Extraction System" document
        by John Swinbank, available from TKP svn.
        """
        # Map our chunks onto a list of islands.

        if labelled_data is None:
            (labels, labelled_data, num_islands, maxposs, maxis, npixs,
             indices, slices) =\
                self.label_islands(detectionthresholdmap,
                                   analysisthresholdmap, deblend_nthresh
                                   )

        start_post_labelling = time.time()

        num_islands = len(labels)

        results = containers.ExtractionResults()

        if deblend_nthresh or force_beam or not VECTORIZED:
            island_list = []
            for label in labels:
                chunk = slices[label - 1]
                analysis_threshold = (analysisthresholdmap[chunk] /
                                      self.rmsmap[chunk]).max()
                # In selected_data only the pixels with the "correct"
                # (see above) labels are retained. Other pixel values are
                # set to -(bignum).
                # In this way, disconnected pixels within (rectangular)
                # slices around islands (particularly the large ones) do
                # not affect the source measurements.
                selected_data = numpy.ma.where(
                    labelled_data[chunk] == label,
                    self.data_bgsubbed[chunk].data, -extract.BIGNUM
                ).filled(fill_value=-extract.BIGNUM)

                island_list.append(
                    extract.Island(
                        selected_data,
                        self.rmsmap[chunk],
                        chunk,
                        analysis_threshold,
                        detectionthresholdmap[chunk],
                        self.beam,
                        deblend_nthresh,
                        DEBLEND_MINCONT,
                        STRUCTURING_ELEMENT
                    )
                )

            if self.islands:
                self.Gaussian_islands = numpy.zeros(self.data.shape)
            # If required, we can save the 'leftovers' from the deblending and
            # fitting processes for later analysis. This needs setting up here:
            if self.residuals:
                self.Gaussian_residuals = numpy.zeros(self.data.shape)
                self.residuals_from_deblending = numpy.zeros(self.data.shape)
                for island in island_list:
                    self.residuals_from_deblending[island.chunk] += (
                        island.data.filled(fill_value=0.))

            # Deblend each of the islands to its consituent parts, if necessary
            if deblend_nthresh:
                deblended_list = [x.deblend() for x in island_list]
                island_list = list(utils.flatten(deblended_list))

            # Set up the fixed fit parameters if 'force beam' is on:
            if force_beam:
                fixed = {'semimajor': self.beam[0],
                         'semiminor': self.beam[1],
                         'theta': self.beam[2]}
            else:
                fixed = None

            # Iterate over the list of islands and measure the source in each,
            # appending it to the results list.
            with Pool(psutil.cpu_count()) as p:
                fit_islands_partial = partial(ImageData.fit_islands,
                                              self.fudge_max_pix_factor,
                                              self.max_pix_variance_factor,
                                              self.beamsize,
                                              self.correlation_lengths, fixed)
                fit_results = p.map(fit_islands_partial, island_list)

            for island, fit_result in zip(island_list, fit_results):
                if fit_result:
                    measurement, Gauss_island, Gauss_residual = fit_result
                else:
                    # Failed to fit; drop this island and go to the next.
                    continue
                try:
                    det = extract.Detection(measurement, self,
                                            chunk=island.chunk, eps_ra=eps_ra,
                                            eps_dec=eps_dec)
                    if (det.ra.error == float('inf') or
                            det.dec.error == float('inf')):
                        logger.warning('Bad fit from blind extraction at pixel coords:'
                                       '%f %f - measurement discarded'
                                       '(increase fitting margin?)', det.x, det.y)
                    else:
                        results.append(det)
                except RuntimeError as e:
                    logger.error("Island not processed; unphysical?")

                if self.islands:
                    self.Gaussian_islands[island.chunk] += Gauss_island

                if self.residuals:
                    self.residuals_from_deblending[island.chunk] -= (
                        island.data.filled(fill_value=0.))
                    self.Gaussian_residuals[island.chunk] += Gauss_residual

        elif num_islands > 0:

            (moments_of_sources, sky_barycenters, xpositions, ypositions,
             ra_errors, dec_errors, error_radii, smaj_asec, errsmaj_asec,
             smin_asec, errsmin_asec, theta_celes_values, theta_celes_errors,
             theta_dc_celes_values, theta_dc_celes_errors) = \
                extract.source_measurements_pixels_and_celestial_vectorised(
                    num_islands, npixs, maxposs, maxis, self.data_bgsubbed.data,
                    self.rmsmap.data, analysisthresholdmap.data, indices,
                    labelled_data, labels, self.wcs, self.fudge_max_pix_factor,
                    self.max_pix_variance_factor,  self.beam, self.beamsize,
                    self.correlation_lengths, eps_ra, eps_dec)

            if self.islands or self.residuals:
                self.Gaussian_islands = numpy.zeros_like(self.data.data)

                # Select the relevant elements of moments_sources, include the
                # peak spectral brightness, but exclude the flux density.
                relevant_moments = (
                    numpy.take(moments_of_sources[:, 0, :],
                               [0, 2, 3, 4, 5, 6], axis=1))

                extract.calculate_Gaussian_islands(indices[:, 0],
                                                   indices[:, 2],
                                                   xpositions, ypositions,
                                                   npixs,
                                                   relevant_moments,
                                                   self.Gaussian_islands)

                if self.residuals:
                    # It only makes sense to compute residuals where we have
                    # reconstructed Gaussian islands, i.e. above the analysis
                    # threshold.
                    # Some parts of self.data_bgsubbed may be masked, but no
                    # sources will have been detected in those masked patches
                    # of the sky, so no need to apply that mask here.
                    self.Gaussian_residuals = \
                        numpy.where(self.Gaussian_islands != 0,
                                    self.data_bgsubbed.data -
                                    self.Gaussian_islands,
                                    0).astype(numpy.float32)

            for count, label in enumerate(labels):
                chunk = slices[label - 1]

                param = extract.ParamSet()
                param.sig = maxis[count] / self.rmsmap.data[tuple(maxposs[count])]

                param["peak"] = Uncertain(moments_of_sources[count, 0, 0], moments_of_sources[count, 1, 0])
                param["flux"] = Uncertain(moments_of_sources[count, 0, 1], moments_of_sources[count, 1, 1])
                param["xbar"] = Uncertain(moments_of_sources[count, 0, 2], moments_of_sources[count, 1, 2])
                param["ybar"] = Uncertain(moments_of_sources[count, 0, 3], moments_of_sources[count, 1, 3])
                param["semimajor"] = Uncertain(moments_of_sources[count, 0, 4], moments_of_sources[count, 1, 4])
                param["semiminor"] = Uncertain(moments_of_sources[count, 0, 5], moments_of_sources[count, 1, 5])
                param["theta"] = Uncertain(moments_of_sources[count, 0, 6], moments_of_sources[count, 1, 6])
                param["semimaj_deconv"] = Uncertain(moments_of_sources[count, 0, 7], moments_of_sources[count, 1, 7])
                param["semimin_deconv"] = Uncertain(moments_of_sources[count, 0, 8], moments_of_sources[count, 1, 8])
                param["theta_deconv"] = Uncertain(moments_of_sources[count, 0, 9], moments_of_sources[count, 1, 9])

                det = extract.Detection(param, self, chunk=chunk)
                results.append(det)

        end_post_labelling = time.time()
        print("Post labelling took {:7.2f} seconds.".format(end_post_labelling-start_post_labelling))

        def is_usable(det):
            # Check that both ends of each axis are usable; that is, that they
            # fall within an unmasked part of the image.
            # The axis will not likely fall exactly on a pixel number, so
            # check all the surroundings.
            def check_point(x, y):
                x = (int(x), int(numpy.ceil(x)))
                y = (int(y), int(numpy.ceil(y)))
                for position in itertools.product(x, y):
                    try:
                        if self.data.mask[position[0], position[1]]:
                            # Point falls in mask
                            return False
                    except IndexError:
                        # Point falls completely outside image
                        return False
                # Point is ok
                return True

            for point in (
                    (det.start_smaj_x, det.start_smaj_y),
                    (det.start_smin_x, det.start_smin_y),
                    (det.end_smaj_x, det.end_smaj_y),
                    (det.end_smin_x, det.end_smin_y)
            ):
                if not check_point(*point):
                    logger.debug("Unphysical source at pixel %f, %f" % (
                    det.x.value, det.y.value))
                    return False
            return True
        # Filter will return a list; ensure we return an ExtractionResults.
        return containers.ExtractionResults(list(filter(is_usable, results)))



import logging
import warnings
from math import degrees

from casacore.tables import table as casacore_table

from sourcefinder.accessors.dataaccessor import DataAccessor
from sourcefinder.utility.coordinates import WCS
from sourcefinder.utility.coordinates import mjd2datetime

logger = logging.getLogger(__name__)


class CasaImage(DataAccessor):
    """
    Provides common functionality for pulling data from the CASA image format

    (Technically known as the 'MeasurementSet' format.)

    NB CasaImage does not provide tau_time or taustart_ts, as there was
    no clear standard for these metadata, so cannot be
    instantiated directly - subclass it and extract these attributes
    as appropriate to a given telescope.
    """

    def __init__(self, url, plane=0, beam=None):
        # super().__init__()
        self.url = url

        # we don't want the table as a property since it makes the accessor
        # not serializable
        table = casacore_table(self.url, ack=False)
        self.data = self.parse_data(table, plane)
        self.wcs = self.parse_coordinates(table)
        self.centre_ra, self.centre_decl = self.parse_phase_centre(table)
        self.freq_eff, self.freq_bw = self.parse_frequency(table)
        self.pixelsize = self.parse_pixelsize()

        if beam:
            (bmaj, bmin, bpa) = beam
        else:
            bmaj, bmin, bpa = self.parse_beam(table)
        self.beam = self.degrees2pixels(
            bmaj, bmin, bpa, self.pixelsize[0], self.pixelsize[1])

    def parse_data(self, table, plane=0):
        """extract and massage data from CASA table"""
        data = table[0]['map'].squeeze()
        planes = len(data.shape)
        if planes != 2:
            msg = "received datacube with %s planes, assuming Stokes I and taking plane 0" % planes
            logger.warning(msg)
            warnings.warn(msg)
            data = data[plane, :, :]
        data = data.transpose()
        return data

    def parse_coordinates(self, table):
        """Returns a WCS object"""
        wcs = WCS()
        my_coordinates = table.getkeyword('coords')['direction0']
        wcs.crval = my_coordinates['crval']
        wcs.crpix = my_coordinates['crpix']
        wcs.cdelt = my_coordinates['cdelt']
        ctype = ['unknown', 'unknown']
        # What about other projections?!
        if my_coordinates['projection'] == "SIN":
            if my_coordinates['axes'][0] == "Right Ascension":
                ctype[0] = "RA---SIN"
            if my_coordinates['axes'][1] == "Declination":
                ctype[1] = "DEC--SIN"
        wcs.ctype = tuple(ctype)
        # Rotation, units? We better set a default
        wcs.crota = (0., 0.)
        wcs.cunit = table.getkeyword('coords')['direction0']['units']
        return wcs

    def parse_frequency(self, table):
        """extract frequency related information from headers"""
        freq_eff = table.getkeywords()['coords']['spectral2']['restfreq']
        freq_bw = table.getkeywords()['coords']['spectral2']['wcs']['cdelt']
        return freq_eff, freq_bw

    def parse_beam(self, table):
        """
        Returns:
          - Beam parameters, (semimajor, semiminor, parallactic angle) in
            (pixels,pixels, radians).
        """

        def ensure_degrees(quantity):
            if quantity['unit'] == 'deg':
                return quantity['value']
            elif quantity['unit'] == 'arcsec':
                return quantity['value'] / 3600
            elif quantity['unit'] == 'rad':
                return degrees(quantity['value'])
            else:
                raise Exception("Beam units (%s) unknown" % quantity['unit'])

        restoringbeam = table.getkeyword('imageinfo')['restoringbeam']
        bmaj = ensure_degrees(restoringbeam['major'])
        bmin = ensure_degrees(restoringbeam['minor'])
        bpa = ensure_degrees(restoringbeam['positionangle'])
        return bmaj, bmin, bpa

    def parse_phase_centre(self, table):
        # The units for the pointing centre are not given in either the image
        # cubeitself or in the ICD. Assume radians.
        # Note that we'll return the RA modulo 360 so it's always 0 <= RA < 360
        centre_ra, centre_decl = table.getkeyword('coords')['pointingcenter'][
            'value']
        return degrees(centre_ra) % 360, degrees(centre_decl)

    def parse_taustartts(self, table):
        """
        Extract integration start-time from CASA table header.

        This applies to some CASA images (typically those created from uvFITS
        files) but not all, and so should be called for each sub-class.

        Arguments:
            table: MAIN table of CASA image.

        Returns:
            Time of image start as a instance of ``datetime.datetime``
        """
        obsdate = table.getkeyword('coords')['obsdate']['m0']['value']
        return mjd2datetime(obsdate)

    @staticmethod
    def unique_column_values(table, column_name):
        """
        Find all the unique values in a particular column of a CASA table.

        Arguments:
          - table:       ``casacore.tables.table``
          - column_name: ``str``

        Returns:
          - ``numpy.ndarray`` containing unique values in column.
        """
        return table.query(
            columns=column_name, sortlist="unique %s" % (column_name)
        ).getcol(column_name)



"""
Mock / synthetic data objects for use in testing.
"""
import numpy as np
from sourcefinder.accessors.dataaccessor import DataAccessor
from sourcefinder.utility.coordinates import WCS
import datetime

class Mock(object):
    def __init__(self, returnvalue=None):
        self.callcount = 0
        self.callvalues = []
        self.returnvalue = returnvalue

    def __call__(self, *args, **kwargs):
        self.callcount += 1
        self.callvalues.append((args, kwargs))
        return self.returnvalue


def make_wcs(crval=None,
             cdelt=None,
             crpix=None
             ):
    """
    Make a WCS object for insertion into a synthetic image.

    Args:
        crval (tuple): Tuple of (RA, Dec) in decimal degrees at the reference
            position.
        crpix (tuple): Tuple of (x,y) co-ordinates describing the reference
            pixel location corresponding to the crval sky-position.
        cdelt (tuple): Tuple of (cdelt0, cdelt1) in decimal degrees.
            This is the pixel width in degrees of arc, but not necessarily
            aligned to RA, Dec unless `crota` is (0,0). If that *is* the case,
            then typically cdelt0 is negative since the x-axis is in direction
            of West (decreasing RA).

    """
    # For any arguments not set we simply assign an arbitrary valid value:
    if crval is None:
        crval = 100., 45.
    if cdelt is None:
        pixel_width_arcsec = 40
        pixel_width_deg = pixel_width_arcsec / 3600.
        cdelt = (-pixel_width_deg, pixel_width_deg)
    if crpix is None:
        crpix = (256.0, 256.0)
    wcs = WCS()
    wcs.cdelt = cdelt
    wcs.crota = (0.0, 0.0)
    wcs.crpix = crpix
    wcs.crval = crval
    wcs.ctype = ('RA---SIN', 'DEC--SIN')
    wcs.cunit = ('deg', 'deg')
    return wcs

class SyntheticImage(DataAccessor):
    def __init__(self,
                 wcs=None,
                 data=None,
                 beam=(1.5,1.5,0),
                 freq_eff=150e6,
                 freq_bw=2e6,
                 tau_time=1800,
                 taustart_ts=datetime.datetime(2015,1,1)
                 ):
        """
        Generate a synthetic image for use in tests

        Args:
            wcs (sourcefinder.utility.coordinates.WCS): WCS for the image.
            data (array_like): Data for the image. Default is a 512x512 array of
                zeroes.
            beam (tuple): Beamsemi-major axis (in pixels), semi-minor axis (pixels)
                and position angle (radians).
            freq_eff(float): Effective frequency of the image in Hz.
                That is, the mean frequency of all the visibility data which
                comprises this image.
            freq_bw(float): The frequency bandwidth of this image in Hz.
            tau_time(float): Total time on sky in seconds.
            taustart_ts(float): Timestamp of the first integration which
                constitutes part of this image. MJD in seconds.
        """
        self.url = "SyntheticImage"
        self.wcs = wcs
        if self.wcs is None:
            self.wcs = make_wcs()
        self.data = data
        if self.data is None:
            self.data = np.zeros((512,512))
        self.beam = beam
        self.freq_eff = freq_eff
        self.freq_bw = freq_bw
        self.tau_time = tau_time
        self.taustart_ts = taustart_ts

        self.pixelsize = self.parse_pixelsize()
        self.centre_ra, self.centre_decl = self.calculate_phase_centre()


    def calculate_phase_centre(self):
        x, y = self.data.shape
        centre_ra, centre_decl = self.wcs.p2s((x / 2, y / 2))
        return centre_ra, centre_decl



"""
Container classes for the TKP pipeline.

These provide convenient means of marshalling the various types of data --
lightcurves, detections, sources, etc -- that the pipeline must handle.
"""

import logging

logger = logging.getLogger(__name__)


class ObjectContainer(list):
    """A container class for objects.

    What sort of objects? Well, anything that has a position and we
    want to keep lists of, really. So detections (ie, an individual
    source measurement on an image), sources (ie all the detections of
    a given object in a given image stack) and lightcurves (ie, all
    the sources associated with a given object through time).

    You probably don't want to use this on it's own: see ExtractionResults,
    TargetList or source for more useful derived classes.
    """

    def closest_to(self, pix_x, pix_y):
        distance, target = False, False
        logger.debug("Beginning a search for objects near %.1f, %.1f: ",
                     pix_x, pix_y)
        logger.debug("%s contains %d objects", str(self), len(self))
        for obj in self:
            tmpdist = (pix_x - obj.x) ** 2 + (pix_y - obj.y) ** 2
            logger.debug("Object at %f, %f", obj.x, obj.y)
            logger.debug("Has distance %f", tmpdist)
            if not distance:
                distance = tmpdist
                target = obj
            else:
                if tmpdist < distance:
                    target = obj
                    distance = tmpdist
            logger.debug("Best distance is now %f", distance)
            logger.debug("From object %s", str(target))
        if not distance:
            return (target, distance)
        else:
            return (target, distance ** 0.5)

    def __setslice__(self, section, items):
        """
        Not implemented.
        """
        raise NotImplementedError

    def __iadd__(self, y):
        """
        Not implemented.
        """
        raise NotImplementedError

    def __imul__(self, y):
        """
        Not implemented.
        """
        raise NotImplementedError

    def __mul__(self, y):
        """
        Not implemented.
        """
        raise NotImplementedError

    def __rmul__(self, y):
        """
        Not implemented.
        """
        raise NotImplementedError

    def __str__(self):
        return 'Container: ' + str(len(self)) + ' object(s).'


class ExtractionResults(ObjectContainer):
    """Container for the results of running source extraction on an
    ImageData object"""

    def __str__(self):
        return 'ExtractionResults: ' + str(len(self)) + ' detection(s).'



"""
This module contain utilities for the source finding routines
"""

import math

import numpy
import scipy.integrate

from sourcefinder.gaussian import gaussian
from sourcefinder.utility import coordinates

def generate_subthresholds(min_value, max_value, num_thresholds):
    """
    Generate a series of ``num_thresholds`` logarithmically spaced values
    in the range (min_value, max_value) (both exclusive).
    """
    # First, we calculate a logarithmically spaced sequence between exp(0.0)
    # and (max - min + 1). That is, the total range is between 1 and one
    # greater than the difference between max and min.
    # We subtract 1 from this to get the range between 0 and (max-min).
    # We add min to that to get the range between min and max.
    subthrrange = numpy.logspace(
        0.0,
        numpy.log(max_value + 1 - min_value),
        num=num_thresholds + 1,  # first value == min_value
        base=numpy.e,
        endpoint=False  # do not include max_value
    )[1:]
    subthrrange += (min_value - 1)
    return subthrrange


def get_error_radius(wcs, x_value, x_error, y_value, y_error):
    """
    Estimate an absolute angular error on the position (x_value, y_value)
    with the given errors.

    This is a pessimistic estimate, because we take sum of the error
    along the X and Y axes. Better might be to project them both back on
    to the major/minor axes of the elliptical fit, but this should do for
    now.
    """
    error_radius = 0.0
    try:
        centre_ra, centre_dec = wcs.p2s([x_value, y_value])
        # We check all possible combinations in case we have a nonlinear
        # WCS.
        for pixpos in [
            (x_value + x_error, y_value + y_error),
            (x_value - x_error, y_value + y_error),
            (x_value + x_error, y_value - y_error),
            (x_value - x_error, y_value - y_error)
        ]:
            error_ra, error_dec = wcs.p2s(pixpos)
            error_radius = max(
                error_radius,
                coordinates.angsep(centre_ra, centre_dec, error_ra, error_dec)
            )
    except RuntimeError:
        # We get a runtime error from wcs.p2s if the errors place the
        # limits outside of the image, in which case we set the angular
        # uncertainty to infinity.
        error_radius = float('inf')
    return error_radius

def circular_mask(xdim, ydim, radius):
    """
    Returns a numpy array of shape (xdim, ydim). All points with radius of
    the centre are set to 0; outside that region, they are set to 1.
    """
    centre_x, centre_y = (xdim - 1) / 2.0, (ydim - 1) / 2.0
    x, y = numpy.ogrid[-centre_x:xdim - centre_x, -centre_y:ydim - centre_y]
    return x * x + y * y >= radius * radius


def generate_result_maps(data, sourcelist):
    """Return a source and residual image

    Given a data array (image) and list of sources, return two images, one
    showing the sources themselves and the other the residual after the
    sources have been removed from the input data.
    """
    residual_map = numpy.array(data)  # array constructor copies by default
    gaussian_map = numpy.zeros(residual_map.shape)
    for src in sourcelist:
        # Include everything with 6 times the std deviation along the major
        # axis. Should be very very close to 100% of the flux.
        box_size = 6 * src.smaj.value / math.sqrt(2 * math.log(2))

        lower_bound_x = max(0, int(src.x.value - 1 - box_size))
        upper_bound_x = min(residual_map.shape[0],
                            int(src.x.value - 1 + box_size))
        lower_bound_y = max(0, int(src.y.value - 1 - box_size))
        upper_bound_y = min(residual_map.shape[1],
                            int(src.y.value - 1 + box_size))

        local_gaussian = gaussian(
            src.peak.value,
            src.x.value,
            src.y.value,
            src.smaj.value,
            src.smin.value,
            src.theta.value
        )(
            numpy.indices(residual_map.shape)[0, lower_bound_x:upper_bound_x,
            lower_bound_y:upper_bound_y],
            numpy.indices(residual_map.shape)[1, lower_bound_x:upper_bound_x,
            lower_bound_y:upper_bound_y]
        )

        gaussian_map[lower_bound_x:upper_bound_x,
        lower_bound_y:upper_bound_y] += local_gaussian
        residual_map[lower_bound_x:upper_bound_x,
        lower_bound_y:upper_bound_y] -= local_gaussian

    return gaussian_map, residual_map


def calculate_correlation_lengths(semimajor, semiminor):
    """Calculate the Condon correlation length

    In order to derive the error bars from Gauss fitting from the
    Condon (1997, PASP 109, 116C) formulae, one needs the so called
    correlation length. The Condon formulae assumes a circular area
    with diameter theta_N (in pixels) for the correlation. This was
    later generalized by Hopkins et al. (2003, AJ 125, 465) for
    correlation areas which are not axisymmetric.

    Basically one has theta_N^2 = theta_B*theta_b.

    Good estimates in general are:

    + theta_B = 2.0 * semimajar

    + theta_b = 2.0 * semiminor
    """

    return (2.0 * semimajor, 2.0 * semiminor)


def calculate_beamsize(semimajor, semiminor):
    """Calculate the beamsize based on the semi major and minor axes"""

    return numpy.pi * semimajor * semiminor


def fudge_max_pix(semimajor, semiminor, theta):
    """Estimate peak flux correction at pixel of maximum flux

    Previously, we adopted Rengelink's correction for the
    underestimate of the peak of the Gaussian by the maximum pixel
    method: fudge_max_pix = 1.06. See the WENSS paper
    (1997A&AS..124..259R) or his thesis.  (The peak of the Gaussian
    is, of course, never at the exact center of the pixel, that's why
    the maximum pixel method will always underestimate it.)

    But, instead of just taking 1.06 one can make an estimate of the
    overall correction by assuming that the true peak is at a random
    position on the peak pixel and averaging over all possible
    corrections.  This overall correction makes use of the beamshape,
    so strictly speaking only accurate for unresolved sources.
    After some investigation, it turns out that this method requires not only
    unresolved sources, but also a circular beam shape. With the general
    elliptical beam, the peak source pixel may be located more than half a
    pixel away from the true peak. In that case the integral below will not
    suffice as a correction. Calculating an overall correction for an ensemble
    of sources will, in the case of an elliptical beam shape, become much
    more involved.
    """

    # scipy.integrate.dblquad: Computes a double integral
    # from the scipy docs:
    #   Return the double (definite) integral of f1(y,x) from x=a..b
    #   and y=f2(x)..f3(x).

    log20 = numpy.log(2.0)
    cos_theta = numpy.cos(theta)
    sin_theta = numpy.sin(theta)

    def landscape(y, x):
        up = math.pow(((cos_theta * x + sin_theta * y) / semiminor), 2)
        down = math.pow(((cos_theta * y - sin_theta * x) / semimajor), 2)
        return numpy.exp(log20 * (up + down))

    (correction, abserr) = scipy.integrate.dblquad(landscape, -0.5, 0.5,
                                                   lambda ymin: -0.5,
                                                   lambda ymax: 0.5)

    return correction


def maximum_pixel_method_variance(semimajor, semiminor, theta):
    """Estimate variance for peak flux at pixel position of maximum

    When we use the maximum pixel method, with a correction
    fudge_max_pix, there should be no bias, unless the peaks of the
    Gaussians are not randomly distributed, but relatively close to
    the centres of the pixels due to selection effects from detection
    thresholds.

    Disregarding the latter effect and noise, we can compute the
    variance of the maximum pixel method by integrating (the true
    flux-the average true flux)^2 = (the true flux-fudge_max_pix)^2
    over the pixel area and dividing by the pixel area ( = 1).  This
    is just equal to integral of the true flux^2 over the pixel area
    - fudge_max_pix^2.
    """

    # scipy.integrate.dblquad: Computes a double integral
    # from the scipy docs:
    #   Return the double (definite) integral of f1(y,x) from x=a..b
    #   and y=f2(x)..f3(x).

    log20 = numpy.log(2.0)
    cos_theta = numpy.cos(theta)
    sin_theta = numpy.sin(theta)

    def landscape(y, x):
        return numpy.exp(2.0 * log20 *
                         (
                         math.pow(((cos_theta * x + sin_theta * y) / semiminor),
                                  2) +
                         math.pow(((cos_theta * y - sin_theta * x) / semimajor),
                                  2)
                         )
                         )

    (result, abserr) = scipy.integrate.dblquad(landscape, -0.5, 0.5,
                                               lambda ymin: -0.5,
                                               lambda ymax: 0.5)
    variance = result - math.pow(fudge_max_pix(semimajor, semiminor, theta), 2)

    return variance


def flatten(nested_list):
    """Flatten a nested list

    Nested lists are made in the deblending algorithm. They're
    awful. This is a piece of code I grabbed from
    http://www.daniweb.com/code/snippet216879.html.

    The output from this method is a generator, so make sure to turn
    it into a list, like this::

        flattened = list(flatten(nested)).
    """
    for elem in nested_list:
        if isinstance(elem, (tuple, list, numpy.ndarray)):
            for i in flatten(elem):
                yield i
        else:
            yield elem



"""
This module implements the CASA kat7 data container format.
"""
import logging

from casacore.tables import table as casacore_table

from sourcefinder.accessors.casaimage import CasaImage

logger = logging.getLogger(__name__)


class Kat7CasaImage(CasaImage):
    """
    Use casacore to pull image data out of a CASA table as produced by KAT-7.

    Note that KAT-7 does not currently include image duration in its headers,
    so we use a placeholder value of 1.

    Note also that the start time is taken from the CASA coords record, and
    may not be valid if the image is composed of multiple observations.

    Args:
      - url: location of CASA table
      - plane: if datacube, what plane to use
      - beam: (optional) beam parameters in degrees, in the form
        (bmaj, bmin, bpa). Will attempt to read from header if
        not supplied.
    """

    def __init__(self, url, plane=0, beam=None):
        super().__init__(url, plane, beam)
        table = casacore_table(self.url, ack=False)
        self.taustart_ts = self.parse_taustartts(table)
        self.tau_time = 1  # Placeholder value



import os
import unittest


def requires_database():
    if os.environ.get("TKP_DISABLEDB", False):
        return unittest.skip("Database functionality disabled in configuration")
    return lambda func: func


def requires_data(*args):
    for filename in args:
        if not os.path.exists(filename):
            return unittest.skip("Test data (%s) not available" % filename)
    return lambda func: func


def requires_module(module_name):
    try:
        __import__(module_name)
    except ImportError:
        return unittest.skip("Required module (%s) not available" % module_name)
    return lambda func: func


def duration(test_duration):
    max_duration = float(os.environ.get("TKP_MAXTESTDURATION", False))
    if max_duration:
        if max_duration < test_duration:
            return unittest.skip(
             "Tests of duration > %s disabled with TKP_MAXTESTDURATION" %
                max_duration)
    return lambda func: func


def requires_test_db_managed():
    """
    This decorator is used to disable tests that do potentially low level
    database management operations like destroy and create. You can enable
    these tests by setting the TKP_TESTDBMANAGEMENT environment variable.
    """
    if os.environ.get('TKP_DBENGINE', 'postgresql') == 'monetdb':
        return unittest.skip("DB management tests not supported for Monetdb,"
                             "must be tested manually.")

    if os.environ.get("TKP_TESTDBMANAGEMENT", False):
        return lambda func: func
    return unittest.skip("DB management tests disabled, TKP_TESTDBMANAGEMENT"
                         " not set")

def high_ram_requirements():
    """
    Used to disable tests that break Travis due to out-of-memory issues.
    """
    if os.environ.get("TRAVIS", False):
        return unittest.skip("High-ram requirement unit-tests disabled on Travis")
    return lambda func: func


from sourcefinder.accessors.fitsimage import FitsImage
from sourcefinder.accessors.lofaraccessor import LofarAccessor


class LofarFitsImage(FitsImage, LofarAccessor):
    def __init__(self, url, plane=False, beam=False, hdu=0):
        super().__init__(url, plane, beam, hdu)
        header = self._get_header(hdu)
        self.antenna_set = header['ANTENNA']
        self.ncore = header['NCORE']
        self.nintl = header['NINTL']
        self.nremote = header['NREMOTE']
        self.subbands = header['SUBBANDS']
        self.subbandwidth = header['SUBBANDW']
        if 'TAU_TIME' in header:
            # This may have been set already by _timeparse, but if defined
            # here it is set by our inject script and should be used
            self.tau_time = header['TAU_TIME']



"""
Definition of a two dimensional elliptical Gaussian.
"""

from numpy import exp, log, cos, sin


def gaussian(height, center_x, center_y, semimajor, semiminor, theta):
    """Return a 2D Gaussian function with the given parameters.

    Args:

        height (float): (z-)value of the 2D Gaussian

        center_x (float): x center of the Gaussian

        center_y (float): y center of the Gaussian

        semimajor (float): major axis of the Gaussian

        semiminor (float): minor axis of the Gaussian

        theta (float): angle of the 2D Gaussian in radians, measured
            between the semi-major and y axes, in counterclockwise
            direction.

    Returns:
        lambda: 2D Gaussian (function of pixel coords ``(x,y)``)
    """

    return lambda x, y: height * exp(
        -log(2.0) * (((cos(theta) * (x - center_x) +
                       sin(theta) * (y - center_y)) /
                      semiminor) ** 2.0 +
                     ((cos(theta) * (y - center_y) -
                       sin(theta) * (x - center_x)) /
                      semimajor) ** 2.))


def jac_gaussian(gaussianargs):
    """Return the Jacobian of a 2D anisotropic Gaussian

    Args:
        gaussianargs. This is a list containing the usual six
        Gaussian parameters:

            height (float): (z-)value of the 2D Gaussian

            center_x (float): x center of the Gaussian

            center_y (float): y center of the Gaussian

            semimajor (float): major axis of the Gaussian

            semiminor (float): minor axis of the Gaussian

            theta (float): angle of the 2D Gaussian in radians, measured
                between the semi-major and y axes, in counterclockwise
                direction.

    Returns:
        function: Jacobian of Gaussian, i.e. the derivatives along each of these
                six parameters of the Gaussian, as a function of pixel
                coordinates x and y, so six functions are returned.

    """
    height, center_x, center_y, semimajor, semiminor, theta = gaussianargs

    # First define some auxiliary quantities, which will return in many of the
    # partial derivatives.
    def a(x):
        return center_x - x

    b = semimajor

    def c(y):
        return center_y - y

    d = semiminor
    e = cos(theta)
    f = sin(theta)
    g = log(2)

    def term3(x, y):
        return e * a(x) + f * c(y)

    def term4(x, y):
        return e * c(y) - f * a(x)

    def term1(x, y):
        return term3(x, y) / d

    def term2(x, y):
        return term4(x, y) / b

    def arg(x, y):
        return g * (term1(x, y) ** 2 + term2(x, y) ** 2)

    def expon(x, y):
        return exp(-arg(x, y))

    def common(x, y):
        return 2 * g * height * expon(x, y)

    # Partial derivative of the Gausian along height.
    def dg_dh(x, y): return expon(x, y)

    # Along center_x
    def dg_dx0(x, y):
        return common(x, y) * (-term1(x, y) * e / d + term2(x, y) * f / b)

    # Along center_y
    def dg_dy0(x, y):
        return common(x, y) * (-term1(x, y) * f / d - term2(x, y) * e / b)

    # Along the semi-major axis
    def dg_dsmaj(x, y):
        return common(x, y) * term2(x, y) ** 2 / b

    # Along the semi-minor axus
    def dg_dsmin(x, y):
        return common(x, y) * term1(x, y) ** 2 / d

    # Along the position angle
    def dg_dtheta(x, y):
        return common(x, y) * term3(x, y) * term4(x, y) * (1 / b ** 2
                                                           - 1 / d ** 2)

    jacobian = {"peak": dg_dh,
                "xbar": dg_dx0,
                "ybar": dg_dy0,
                "semimajor": dg_dsmaj,
                "semiminor": dg_dsmin,
                "theta": dg_dtheta}

    return jacobian



#
# LOFAR Transients Key Project
"""
General purpose astronomical coordinate handling routines.
"""
import numpy
import datetime
import logging
import math
import sys
from numba import njit, guvectorize, float64

import pytz
from astropy import wcs as pywcs
from casacore.measures import measures
from casacore.quanta import quantity

logger = logging.getLogger(__name__)

# Note that we take a +ve longitude as WEST.
CORE_LAT = 52.9088
CORE_LON = -6.8689

# ITRF position of CS002
# Should be a good approximation for anything refering to the LOFAR core.
ITRF_X = 3826577.066110000
ITRF_Y = 461022.947639000
ITRF_Z = 5064892.786

# Useful constants
SECONDS_IN_HOUR = 60 ** 2
SECONDS_IN_DAY = 24 * SECONDS_IN_HOUR


def julian_date(time=None, modified=False):
    """
    Calculate the Julian date at a given timestamp.



    Args:
        time (datetime.datetime): Timestamp to calculate JD for.
        modified (bool): If True, return the Modified Julian Date:
            the number of days (including fractions) which have elapsed between
            the start of 17 November 1858 AD and the specified time.
    Returns:
        float: Julian date value.
    """
    if not time:
        time = datetime.datetime.now(pytz.utc)
    mjdstart = datetime.datetime(1858, 11, 17, tzinfo=pytz.utc)
    mjd = time - mjdstart
    mjd_daynumber = (mjd.days + mjd.seconds / (24. * 60 ** 2) +
                     mjd.microseconds / (24. * 60 ** 2 * 1000 ** 2))
    if modified:
        return mjd_daynumber
    return 2400000.5 + mjd_daynumber


def mjd2datetime(mjd):
    """
    Convert a Modified Julian Date to datetime via 'unix time' representation.

    NB 'unix time' is defined by the casacore/casacore package.
    """
    q = quantity("%sd" % mjd)
    return datetime.datetime.fromtimestamp(q.to_unix_time())


def mjd2lst(mjd, position=None):
    """
    Converts a Modified Julian Date into Local Apparent Sidereal Time in
    seconds at a given position. If position is None, we default to the
    reference position of CS002.

    mjd -- Modified Julian Date (float, in days)
    position -- Position (casacore measure)
    """
    dm = measures()
    position = position or dm.position(
        "ITRF", "%fm" % ITRF_X, "%fm" % ITRF_Y, "%fm" % ITRF_Z
    )
    dm.do_frame(position)
    last = dm.measure(dm.epoch("UTC", "%fd" % mjd), "LAST")
    fractional_day = last['m0']['value'] % 1
    return fractional_day * 24 * SECONDS_IN_HOUR


def mjds2lst(mjds, position=None):
    """
    As mjd2lst(), but takes an argument in seconds rather than days.

    Args:
        mjds (float):Modified Julian Date (in seconds)
        position (casacore measure): Position for LST calcs
    """
    return mjd2lst(mjds / SECONDS_IN_DAY, position)


def jd2lst(jd, position=None):
    """
    Converts a Julian Date into Local Apparent Sidereal Time in seconds at a
    given position. If position is None, we default to the reference position
    of CS002.

    Args:
        jd (float): Julian Date
        position (casacore measure): Position for LST calcs.
    """
    return mjd2lst(jd - 2400000.5, position)


# NB: datetime is not sensitive to leap seconds.
# However, leap seconds were first introduced in 1972.
# So there are no leap seconds between the start of the
# Modified Julian epoch and the start of the Unix epoch,
# so this calculation is safe.
# julian_epoch = datetime.datetime(1858, 11, 17)
# unix_epoch = datetime.datetime(1970, 1, 1, 0, 0)
# delta = unix_epoch - julian_epoch
# deltaseconds = delta.total_seconds()
# unix_epoch = 3506716800

# The above is equivalent to this:
unix_epoch = quantity("1970-01-01T00:00:00").get_value('s')


def julian2unix(timestamp):
    """
    Convert a modifed julian timestamp (number of seconds since 17 November
    1858) to Unix timestamp (number of seconds since 1 January 1970).

    Args:
        timestamp (numbers.Number): Number of seconds since the Unix epoch.

    Returns:
        numbers.Number: Number of seconds since the modified Julian epoch.
    """
    return timestamp - unix_epoch


def unix2julian(timestamp):
    """
    Convert a Unix timestamp (number of seconds since 1 January 1970) to a
    modified Julian timestamp (number of seconds since 17 November 1858).

    Args:
        timestamp (numbers.Number): Number of seconds since the modified
            Julian epoch.

    Returns:
        numbers.Number: Number of seconds since the Unix epoch.
    """
    return timestamp + unix_epoch


def sec2deg(seconds):
    """Seconds of time to degrees of arc"""
    return 15.0 * seconds / 3600.0


def sec2days(seconds):
    """Seconds to number of days"""
    return seconds / (24.0 * 3600)


def sec2hms(seconds):
    """Seconds to hours, minutes, seconds"""
    hours, seconds = divmod(seconds, 60 ** 2)
    minutes, seconds = divmod(seconds, 60)
    return (int(hours), int(minutes), seconds)


def altaz(mjds, ra, dec, lat=CORE_LAT):
    """Calculates the azimuth and elevation of source from time and position
    on sky.  Takes MJD in seconds and ra, dec in degrees.  Returns (alt, az) in
    degrees."""

    # compute hour angle in degrees
    ha = mjds2lst(mjds) - ra
    if (ha < 0):
        ha = ha + 360

    # convert degrees to radians
    ha, dec, lat = [math.radians(value) for value in (ha, dec, lat)]

    # compute altitude in radians
    sin_alt = (math.sin(dec) * math.sin(lat) +
               math.cos(dec) * math.cos(lat) * math.cos(ha))
    alt = math.asin(sin_alt)

    # compute azimuth in radians
    # divide by zero error at poles or if alt = 90 deg
    cos_az = ((math.sin(dec) - math.sin(alt) * math.sin(lat)) /
              (math.cos(alt) * math.cos(lat)))
    az = math.acos(cos_az)
    # convert radians to degrees
    hrz_altitude, hrz_azimuth = [math.degrees(value) for value in (alt, az)]
    # choose hemisphere
    if (math.sin(ha) > 0):
        hrz_azimuth = 360 - hrz_azimuth

    return hrz_altitude, hrz_azimuth


def ratohms(radegs):
    """Convert RA in decimal degrees format to hours, minutes,
    seconds format.

    Keyword arguments:
    radegs -- RA in degrees format

    Return value:
    ra -- tuple of 3 values, [hours,minutes,seconds]

    """

    radegs %= 360
    raseconds = radegs * 3600 / 15.0
    return sec2hms(raseconds)


def dectodms(decdegs):
    """Convert Declination in decimal degrees format to hours, minutes,
    seconds format.

    Keyword arguments:
    decdegs -- Dec. in degrees format

    Return value:
    dec -- list of 3 values, [degrees,minutes,seconds]

    """

    sign = -1 if decdegs < 0 else 1
    decdegs = abs(decdegs)
    if decdegs > 90:
        raise ValueError("coordinate out of range")
    decd = int(decdegs)
    decm = int((decdegs - decd) * 60)
    decs = (((decdegs - decd) * 60) - decm) * 60
    # Necessary because of potential roundoff errors
    if decs - 60 > -1e-7:
        decm += 1
        decs = 0
        if decm == 60:
            decd += 1
            decm = 0
            if decd > 90:
                raise ValueError("coordinate out of range")

    if sign == -1:
        if decd == 0:
            if decm == 0:
                decs = -decs
            else:
                decm = -decm
        else:
            decd = -decd
    return (decd, decm, decs)


def propagate_sign(val1, val2, val3):
    """
    casacore (reasonably enough) demands that a minus sign (if required)
    comes at the start of the quantity. Thus "-0D30M" rather than "0D-30M".
    Python regards "-0" as equal to "0"; we need to split off a separate sign
    field.

    If more than one of our inputs is negative, it's not clear what the user
    meant: we raise.

    Args:
        val1(float): (,val2,val3) input values (hour/min/sec or deg/min/sec)

    Returns:
        tuple: "+" or "-" string denoting sign,
            val1, val2, val3 (numeric) denoting absolute values of inputs.
    """
    signs = [x < 0 for x in (val1, val2, val3)]
    if signs.count(True) == 0:
        sign = "+"
    elif signs.count(True) == 1:
        sign, val1, val2, val3 = "-", abs(val1), abs(val2), abs(val3)
    else:
        raise ValueError("Too many negative coordinates")
    return sign, val1, val2, val3


def hmstora(rah, ram, ras):
    """Convert RA in hours, minutes, seconds format to decimal
    degrees format.

    Keyword arguments:
    rah,ram,ras -- RA values (h,m,s)

    Return value:
    radegs -- RA in decimal degrees

    """
    sign, rah, ram, ras = propagate_sign(rah, ram, ras)
    ra = quantity("%s%dH%dM%f" % (sign, rah, ram, ras)).get_value()
    if abs(ra) >= 360:
        raise ValueError("coordinates out of range")
    return ra


def dmstodec(decd, decm, decs):
    """Convert Dec in degrees, minutes, seconds format to decimal
    degrees format.

    Keyword arguments:
    decd, decm, decs -- list of Dec values (d,m,s)

    Return value:
    decdegs -- Dec in decimal degrees

    """
    sign, decd, decm, decs = propagate_sign(decd, decm, decs)
    dec = quantity("%s%dD%dM%f" % (sign, decd, decm, decs)).get_value()
    if abs(dec) > 90:
        raise ValueError("coordinates out of range")
    return dec

def cmp(a, b):
    return bool(a > b) - bool(a < b)

def angsep(ra1, dec1, ra2, dec2):
    """Find the angular separation of two sources, in arcseconds,
    using the proper spherical trig formula

    Keyword arguments:
    ra1,dec1 - RA and Dec of the first source, in decimal degrees
    ra2,dec2 - RA and Dec of the second source, in decimal degrees

    Return value:
    angsep - Angular separation, in arcseconds

    """

    b = (math.pi / 2) - math.radians(dec1)
    c = (math.pi / 2) - math.radians(dec2)
    temp = (math.cos(b) * math.cos(c)) + (
    math.sin(b) * math.sin(c) * math.cos(math.radians(ra1 - ra2)))

    # Truncate the value of temp at +- 1: it makes no sense to do math.acos()
    # of a value outside this range, but occasionally we might get one due to
    # rounding errors.
    if abs(temp) > 1.0:
        temp = 1.0 * cmp(temp, 0)

    return 3600 * math.degrees(math.acos(temp))

@njit
def cmp_jitted(a, b):
    return bool(a > b) - bool(a < b)

@guvectorize([(float64[:], float64[:], float64)],
             '(n), (n) -> ()', nopython=True)
def angsep_vectorized(ra_dec1, ra_dec2, angular_separation):
    """Find the angular separation of two sources, in arcseconds,
    using the proper spherical trig formula

    Positional arguments:
        ra_dec1 (ndarray)- RA and Dec of the first source, in decimal degrees
        ra_dec2 (ndarray)- RA and Dec of the second source, in decimal degrees

    Return value: None, because of the guvectorize decorator, but
        angular_separation (float, arcseconds) is assigned a value.

    """
    ra1, dec1 = ra_dec1
    ra2, dec2 = ra_dec2

    b = (math.pi / 2) - math.radians(dec1)
    c = (math.pi / 2) - math.radians(dec2)
    temp = (math.cos(b) * math.cos(c)) + (
            math.sin(b) * math.sin(c) * math.cos(math.radians(ra1 - ra2)))

    # Truncate the value of temp at +- 1: it makes no sense to do math.acos()
    # of a value outside this range, but occasionally we might get one due to
    # rounding errors.
    if abs(temp) > 1.0:
        temp = 1.0 * cmp_jitted(temp, 0)

    angular_separation = 3600 * math.degrees(math.acos(temp))


def alphasep(ra1, ra2, dec1, dec2):
    """Find the angular separation of two sources in RA, in arcseconds

    Keyword arguments:
    ra1,dec1 - RA and Dec of the first source, in decimal degrees
    ra2,dec2 - RA and Dec of the second source, in decimal degrees

    Return value:
    angsep - Angular separation, in arcseconds

    """

    return 3600 * (ra1 - ra2) * math.cos(math.radians((dec1 + dec2) / 2.0))


def deltasep(dec1, dec2):
    """Find the angular separation of two sources in Dec, in arcseconds

    Keyword arguments:
    dec1 - Dec of the first source, in decimal degrees
    dec2 - Dec of the second source, in decimal degrees

    Return value:
    angsep - Angular separation, in arcseconds

    """

    return 3600 * (dec1 - dec2)


# Find angular separation in Dec of 2 positions, in arcseconds
def alpha(l, m, alpha0, delta0):
    """Convert a coordinate in l,m into an coordinate in RA

    Keyword arguments:
    l,m -- direction cosines, given by (offset in cells) x cellsi (radians)
    alpha_0, delta_0 -- centre of the field

    Return value:
    alpha -- RA in decimal degrees
    """
    return (alpha0 + (math.degrees(math.atan2(l, (
        (math.sqrt(1 - (l * l) - (m * m)) * math.cos(math.radians(delta0))) -
        (m * math.sin(math.radians(delta0))))))))


def alpha_inflate(theta, decl):
    """Compute the ra expansion for a given theta at a given declination
    
    Keyword arguments:
    theta, decl are both in decimal degrees.
    
    Return value:
    alpha -- RA inflation in decimal degrees

    For a derivation, see MSR TR 2006 52, Section 2.1
    http://research.microsoft.com/apps/pubs/default.aspx?id=64524

    """
    if abs(decl) + theta > 89.9:
        return 180.0
    else:
        return math.degrees(abs(math.atan(
            math.sin(math.radians(theta)) / math.sqrt(abs(
                math.cos(math.radians(decl - theta)) * math.cos(
                    math.radians(decl + theta)))))))


# Find the RA of a point in a radio image, given l,m and field centre
def delta(l, m, delta0):
    """Convert a coordinate in l, m into an coordinate in Dec

    Keyword arguments:
    l, m -- direction cosines, given by (offset in cells) x cellsi (radians)
    alpha_0, delta_0 -- centre of the field

    Return value:
    delta -- Dec in decimal degrees
    """
    return math.degrees(math.asin(m * math.cos(math.radians(delta0)) +
                                  (math.sqrt(1 - (l * l) - (m * m)) *
                                   math.sin(math.radians(delta0)))))


def l(ra, dec, cra, incr):
    """Convert a coordinate in RA,Dec into a direction cosine l

    Keyword arguments:
    ra,dec -- Source location
    cra -- RA centre of the field
    incr -- number of degrees per pixel (negative in the case of RA)

    Return value:
    l -- Direction cosine

    """
    return ((math.cos(math.radians(dec)) * math.sin(math.radians(ra - cra))) /
            (math.radians(incr)))


def m(ra, dec, cra, cdec, incr):
    """Convert a coordinate in RA,Dec into a direction cosine m

    Keyword arguments:
    ra,dec -- Source location
    cra,cdec -- centre of the field
    incr -- number of degrees per pixel

    Return value:
    m -- direction cosine

    """
    return ((math.sin(math.radians(dec)) * math.cos(math.radians(cdec))) -
            (math.cos(math.radians(dec)) * math.sin(math.radians(cdec)) *
             math.cos(math.radians(ra - cra)))) / math.radians(incr)


def lm_to_radec(ra0, dec0, l, m):
    """
    Find the l direction cosine in a radio image, given an RA and Dec and the
    field centre
    """
    # This function should be the inverse of radec_to_lmn, but it is
    # not. There is likely an error here.

    sind0 = math.sin(dec0)
    cosd0 = math.cos(dec0)
    dl = l
    dm = m
    d0 = dm * dm * sind0 * sind0 + dl * dl - 2 * dm * cosd0 * sind0
    sind = math.sqrt(abs(sind0 * sind0 - d0))
    cosd = math.sqrt(abs(cosd0 * cosd0 + d0))
    if (sind0 > 0):
        sind = abs(sind)
    else:
        sind = -abs(sind)

    dec = math.atan2(sind, cosd)

    if l != 0:
        ra = math.atan2(-dl, (cosd0 - dm * sind0)) + ra0
    else:
        ra = math.atan2((1e-10), (cosd0 - dm * sind0)) + ra0

        # Calculate RA,Dec from l,m and phase center.  Note: As done in
        # Meqtrees, which seems to differ from l, m functions above.  Meqtrees
        # equation may have problems, judging from my difficulty fitting a
        # fringe to L4086 data.  Pandey's equation is now used in radec_to_lmn

    return (ra, dec)


def radec_to_lmn(ra0, dec0, ra, dec):
    l = math.cos(dec) * math.sin(ra - ra0)
    sind0 = math.sin(dec0)
    if sind0 != 0:
        # from pandey;  gives same results for casa and cyga
        m = (math.sin(dec) * math.cos(dec0) -
             math.cos(dec) * math.sin(dec0) * math.cos(ra - ra0))
    else:
        m = 0
    n = math.sqrt(1 - l ** 2 - m ** 2)
    return (l, m, n)


def eq_to_gal(ra, dec):
    """Find the Galactic co-ordinates of a source given the equatorial
    co-ordinates

    Keyword arguments:
    (alpha,delta) -- RA, Dec in decimal degrees

    Return value:
    (l,b) -- Galactic longitude and latitude, in decimal degrees

    """
    dm = measures()

    result = dm.measure(
        dm.direction("J200", "%fdeg" % ra, "%fdeg" % dec),
        "GALACTIC"
    )
    lon_l = math.degrees(result['m0']['value']) % 360  # 0 < ra < 360
    lat_b = math.degrees(result['m1']['value'])

    return lon_l, lat_b


def gal_to_eq(lon_l, lat_b):
    """Find the Galactic co-ordinates of a source given the equatorial
    co-ordinates

    Keyword arguments:
    (l, b) -- Galactic longitude and latitude, in decimal degrees

    Return value:
    (alpha, delta) -- RA, Dec in decimal degrees

    """
    dm = measures()

    result = dm.measure(
        dm.direction("GALACTIC", "%fdeg" % lon_l, "%fdeg" % lat_b),
        "J2000"
    )
    ra = math.degrees(result['m0']['value']) % 360  # 0 < ra < 360
    dec = math.degrees(result['m1']['value'])

    return ra, dec


def eq_to_cart(ra, dec):
    """Find the cartesian co-ordinates on the unit sphere given the eq. co-ords.

        ra, dec should be in degrees.
    """
    return (
    math.cos(math.radians(dec)) * math.cos(math.radians(ra)),  # Cartesian x
    math.cos(math.radians(dec)) * math.sin(math.radians(ra)),  # Cartesian y
    math.sin(math.radians(dec)))  # Cartesian z


class CoordSystem:
    """A container for constant strings representing different coordinate
    systems."""
    FK4 = "B1950 (FK4)"
    FK5 = "J2000 (FK5)"


def coordsystem(name):
    """Given a string, return a constant from class CoordSystem."""
    mappings = {
        'j2000': CoordSystem.FK5,
        'fk5': CoordSystem.FK5,
        CoordSystem.FK5.lower(): CoordSystem.FK5,
        'b1950': CoordSystem.FK4,
        'fk4': CoordSystem.FK4,
        CoordSystem.FK4.lower(): CoordSystem.FK4
    }
    return mappings[name.lower()]


def convert_coordsystem(ra, dec, insys, outsys):
    """
    Convert RA & dec (given in decimal degrees) between equinoxes.
    """
    dm = measures()

    if insys == CoordSystem.FK4:
        insys = "B1950"
    elif insys == CoordSystem.FK5:
        insys = "J2000"
    else:
        raise Exception("Unknown Coordinate System")

    if outsys == CoordSystem.FK4:
        outsys = "B1950"
    elif outsys == CoordSystem.FK5:
        outsys = "J2000"
    else:
        raise Exception("Unknown Coordinate System")

    result = dm.measure(
        dm.direction(insys, "%fdeg" % ra, "%fdeg" % dec),
        outsys
    )

    ra = math.degrees(result['m0']['value']) % 360  # 0 < ra < 360
    dec = math.degrees(result['m1']['value'])

    return ra, dec


class WCS:
    """
    Wrapper around pywcs.WCS.

    This is primarily to preserve API compatibility with the earlier,
    home-brewed python-wcslib wrapper. It includes:

      * A fix for the reference pixel lying at the zenith;
      * Raises ValueError if coordinates are invalid.
    """
    # ORIGIN is the upper-left corner of the image. pywcs supports both 0
    # (NumPy, C-style) or 1 (FITS, Fortran-style). The TraP uses 1.
    ORIGIN = 1

    # We can set these attributes on the pywcs.WCS().wcs object to configure
    # the coordinate system.
    WCS_ATTRS = ("crpix", "cdelt", "crval", "ctype", "cunit", "crota")

    def __init__(self):
        # Currently, we only support two dimensional images.
        self.wcs = pywcs.WCS(naxis=2)

    def __setattr__(self, attrname, value):
        if attrname in self.WCS_ATTRS:
            # Account for arbitrary coordinate rotations in images pointing at
            # the North Celestial Pole. We set the reference direction to
            # infintesimally less than 90 degrees to avoid any ambiguity. See
            # discussion at #4599.
            if attrname == "crval" and (
                    value[1] == 90 or value[1] == math.pi / 2):
                value = (value[0], value[1] * (1 - sys.float_info.epsilon))
            self.wcs.wcs.__setattr__(attrname, value)
        else:
            super().__setattr__(attrname, value)

    def __getattr__(self, attrname):
        if attrname in self.WCS_ATTRS:
            return getattr(self.wcs.wcs, attrname)
        raise AttributeError(f"{type(self)!r} object has no attribute {attrname!r}")

    def p2s(self, pixpos):
        """
        Pixel to spatial coordinate conversion.

        Args:
            pixpos [list]:  [x, y] pixel position.

        Returns:
            tuple: ra (float) Right ascension corresponding to position [x, y]
                   dec (float) Declination corresponding to position [x, y]
        """
        ra, dec = self.wcs.wcs_pix2world(pixpos[0], pixpos[1], self.ORIGIN)
        if math.isnan(ra) or math.isnan(dec):
            raise RuntimeError("Spatial position is not a number")
        return float(ra), float(dec)

    def s2p(self, spatialpos):
        """
        Spatial to Pixel coordinate conversion.

        Args:
            pixpos (tuple):  [ra, dec] spatial position

        Returns:
            tuple: X pixel value corresponding to position [ra, dec],
                   Y pixel value corresponding to position [ra, dec]
        """
        x, y = self.wcs.wcs_world2pix(spatialpos[0], spatialpos[1], self.ORIGIN)
        if math.isnan(x) or math.isnan(y):
            raise RuntimeError("Pixel position is not a number")
        return float(x), float(y)

    def all_p2s(self, array_of_pixpos):
        """
        Vectorized pixel to spatial coordinate conversion, making use of
        all_pix2world from astropy. This will save time when thousands of
        sources are detected.

        Args:
            array_of_pixpos ((N, 2) array):  array of [x, y] pixel
            positions.

        Returns:
            (N, 2) array: each row has an entry for Right ascension (float)
                          and Declination (float)

        """
        sky_coordinates = self.wcs.all_pix2world(array_of_pixpos, self.ORIGIN)
        if numpy.isnan(sky_coordinates).any():
            raise RuntimeError("Spatial position is not a number")
        # Mimic conditional from extract.Detection._physical_coordinates
        if numpy.any(numpy.abs(abs(sky_coordinates[:, 1] > 90.0))):
            raise ValueError("At least one object falls outside the sky")

        return sky_coordinates



import numpy

from sourcefinder.accessors.dataaccessor import DataAccessor


class LofarHdf5Image(DataAccessor):
    # Not currently instantiable; LOFAR HDF5 images are not in use
    def __init__(self, source, plane=False, beam=False):

        self.plane = plane

        self._beamsizeparse(source)
        self._read_data(source)

        coordinfo = source["/coordinfo"]
        attrgroups = source["/ATTRGROUPS"]

        lofar_antenna = attrgroups["LOFAR_ANTENNA"]
        lofar_field = attrgroups['LOFAR_FIELD']
        lofar_history = attrgroups['LOFAR_HISTORY']
        lofar_observation = attrgroups['LOFAR_OBSERVATION']
        lofar_origin = attrgroups['LOFAR_ORIGIN']
        lofar_source = attrgroups['LOFAR_SOURCE']
        lofar_station = attrgroups['LOFAR_STATION']

        self.url = source.attrs['FILENAME']
        filedate = source.attrs['FILEDATE']
        filetype = source.attrs['FILETYPE']
        self.telescope = source.attrs['TELESCOPE']
        observer = source.attrs['OBSERVER']
        project_id = source.attrs['PROJECT_ID']
        project_title = source.attrs['PROJECT_TITLE'],
        project_pi = source.attrs['PROJECT_PI']
        project_co_i = source.attrs['PROJECT_CO_I']
        project_contact = source.attrs['PROJECT_CONTACT']
        observation_id = source.attrs['OBSERVATION_ID']
        observation_start_mjd = source.attrs['OBSERVATION_START_MJD']
        observation_start_utc = source.attrs['OBSERVATION_START_UTC']
        observation_end_mjd = source.attrs['OBSERVATION_END_MJD']
        observation_end_utc = source.attrs['OBSERVATION_END_UTC']
        observation_nof_station = source.attrs['OBSERVATION_NOF_STATIONS']
        observation_stations_list = source.attrs['OBSERVATION_STATIONS_LIST']
        observation_frequency_max = source.attrs['OBSERVATION_FREQUENCY_MAX']
        observation_frequency_min = source.attrs['OBSERVATION_FREQUENCY_MIN']
        self.freq_eff = source.attrs['OBSERVATION_FREQUENCY_CENTER']
        observation_frequency_unit = source.attrs['OBSERVATION_FREQUENCY_UNIT']
        observation_nof_bits_per_sample = source.attrs[
            'OBSERVATION_NOF_BITS_PER_SAMPLE']
        clock_frequency = source.attrs['CLOCK_FREQUENCY']
        clock_frequency_unit = source.attrs['CLOCK_FREQUENCY_UNIT']
        antenna_set = source.attrs['ANTENNA_SET']
        filter_selection = source.attrs['FILTER_SELECTION']
        target = source.attrs['TARGET']
        system_version = source.attrs['SYSTEM_VERSION']
        pipeline_name = source.attrs['PIPELINE_NAME']
        pipeline_version = source.attrs['PIPELINE_VERSION']
        notes = source.attrs['NOTES']

    def get_header(self):
        # Preserved for API compatibility.
        return self.header

    def _read_data(self, source):
        self.data = numpy.squeeze(source["map"])

        if len(self.data.shape) != 2:
            raise IndexError("Data has wrong shape")

            # TODO: do we need to transpose?
            # self.data = data.transpose()

    def _beamsizeparse(self, source):
        """Read and return the beam properties bmaj, bmin and bpa values from
        the fits header
        """
        # todo: this data is not present in the h5 lofar example file in the data repo
        pass
